(***********************************************************************
 *
 * MathParser.pas version 1.1
 *
 * Copyright (c) 2013,2014 Michel Deslierres
 * All rights reserved. Tous droits rservs.
 *
 * Read full license in the file license.txt
 * Lire la licence au complet dans le fichier license.txt
 *
 ************************************************************************
 *
 * Changements / Changes
 *
 * - version 1.1
 *     Added boolean expressions, boolean operators and the
 *     internal function if.
 *
 ************************************************************************)

(*:
   Unit that performs lexical analysis of simple mathematical
   expressions and evaluates them. This is accomplished in a
   two pass recursive scanner. In the first pass a tree representation
   of the expression is constructed. In the second pass the
   tree is evaluated.

   The only supported operators and their precedence are

     op   operation       precedence
     ()   parenthesis        6
     -    negation           5
     !    not                5
     ^    exponentiation     4     ( also ** )
     *    multiplication     3
     /    division           3
     %    modulus            3
     :    integer division   3
     &    and                3     ( also && )
     +    addition           2
     -    subtraction        2
     |    or                 2     ( also || )
     <    less               1
     <=   less or equal      1
     >    greater            1
     >=   greater or equal   1
     =    equal              1     ( also == )
     <>   not equal          1     ( also != )

WSN definitions of the supported syntax

DecSym  = ? locale dependant character, usually "." in English, "," in French... ?
ListSym = ? locale dependant character, usually "," in English, ";" in French... ?

Digit      = "0" | ... | "9".
Integer    = Digit {Digit}.
Number     = Integer [ DecSym [Integer] [ ("E" | "e") ["-" | "+"] Integer] ].
Letter     = "a" | ... | "z" | "A" | ... | "Z".
Identifier = Letter { Letter | Digit }.

{$IFDEF EXPONENT_RIGHT_ASSOCIATIVE}
Exponent   = Factor { ("^" | "**") Exponent }.
{$ELSE}
Exponent   = Factor { ("^" | "**") Factor }.
{$ENDIF}
Term       = Exponent { ("*" | "/" | ":" | "%" | "&" | "&&" ) Exponent }.
Sum        = Term { ("-" | "+" | "|" ) Term }.
Expression = Sum, [ ("<" | "<=" | "=" | "==" | '>' | '>=' | '<>' | '!='), Sum ].
Params     = "(" Expression { ListSym, Expression } ")".
Call       = Identifier [Params].
Factor     = Number | "(" Expression ")" | Call | ("-" | "+" | "!") Factor.

ParamList  = "(", Identifer { ListSym, Identifer } ")";
Assignment = Identifier, [ParamList], ":=", Expression;

Alternate Expression which would allow for x < y != z:

Expression2 = Sum { ("<" | "<=" | "=" | "==" | "!=" | "<>" | ">=" | ">") Sum }.

Alternat definition of Number 

Integer    = Digit {Digit}.
DecNumber  = Integer [ DecSym {Digit} ]
Number     = DecNumber [ ("E" | "e") ["-" | "+"] Integer ].
*)
Unit MathParser;

{$IFNDEF PARSER_INC}
  {$I Parser.inc}
{$ENDIF}

interface

uses
  SysUtils, ParserTypes;

{: The procedure that performs a lexical analysis of a mathematical expression
   and returns the result as a tree. Can raise an exception of type
   EMathParseErr.}
function Parse(const source: string): TTree;

{: The procedure that performs a lexical analysis of an assignment statement of
   the form "f(a, b) := (a/b)^2". The user defined constant or function is
   added to ExternalTargets as defined in Targets.pas.}
procedure ParseAssign(const source: string);

{: The function that performs a lexical analysis of a mathematical expression
   and returns the result of the calculation. Can raise an exception of type
   EMathParseErr or EMathError (EInvalidOp, EZeroDivide, EOverflow and 
   EUnderflow).}
function EvaluateExpression(const source: string): float;

{: Character used as decimal symbol. By default this is defined in the
   computer's locale (usually a period for English language systems
   and a comma for French language systems).}
function ParserDecimalChar: char;

{: Sets the character to be used as the decimal symbol. Returns the
   previous decimal symbol.}
function SetParserDecimalChar(value: char): char;

{: Character used as list separator symbol. By default this is defined in the
   computer's locale (usually a comma for English language systems
   and a semi-colon for French language systems).}
function ParserListChar: char;

{: Sets the character to be used as the list separator symbol. Returns the
   previous list separator symbol.}
function SetParserListChar(value: char): char;

implementation

uses
  Classes, {$IFDEF HAS_SYSCONST} SysConst, {$ENDIF}
  ParserMath, {$IFDEF USE_MATH_UNIT} Math, {$ENDIF}
  ParserConsts, SimpleTokenizer, Targets;


{$IFDEF DUMP}
const
  NL = #13#10; {new line CR LF}

var
  FLevel: integer = 0;

procedure IncLevel;
begin
  inc(FLevel, 1);
end;

procedure DecLevel;
begin
  dec(FLevel, 1);
end;
{$ENDIF}

{-----------------------------------------------------------}
{$I Nodes.inc}
{-----------------------------------------------------------}

type
  { Term nodes }

  TAndNode = class(TBinaryNode)
  public
    function Eval: float; override;
    {$IFDEF DUMP}
    function ToText: string; override;
    {$ENDIF}
  end;
  
  { Sum nodes }

  TOrNode = class(TBinaryNode)
  public
    function Eval: float; override;
    {$IFDEF DUMP}
    function ToText: string; override;
    {$ENDIF}
  end;

  { Expression nodes }

  TLessNode = class(TBinaryNode)
  public
    function Eval: float; override;
    {$IFDEF DUMP}
    function ToText: string; override;
    {$ENDIF}
  end;

  TLessOrEqualNode = class(TBinaryNode)
  public
    function Eval: float; override;
    {$IFDEF DUMP}
    function ToText: string; override;
    {$ENDIF}
  end;

  TGreaterNode = class(TBinaryNode)
  public
    function Eval: float; override;
    {$IFDEF DUMP}
    function ToText: string; override;
    {$ENDIF}
  end;

  TGreaterOrEqualNode = class(TBinaryNode)
  public
    function Eval: float; override;
    {$IFDEF DUMP}
    function ToText: string; override;
    {$ENDIF}
  end;

  TEqualNode = class(TBinaryNode)
  public
    function Eval: float; override;
    {$IFDEF DUMP}
    function ToText: string; override;
    {$ENDIF}
  end;

  TNotEqualNode = class(TBinaryNode)
  public
    function Eval: float; override;
    {$IFDEF DUMP}
    function ToText: string; override;
    {$ENDIF}
  end;

  { Factor nodes }

  TNotNode = class(TUnaryNode)
  public
    function Eval: float; override;
    {$IFDEF DUMP}
    function ToText: string; override;
    {$ENDIF}
  end;

    {Special call node for the internal if function defined
       if(condition, true_expression, false_expression).
     The standard TCallNode's Eval method will evaluate
     condition, true_expression and false_expression before
     calling an evaluator to decide which value to return.
     This is not acceptable, only one of true_expression
     or false_expression should be evaluated. }
  TIfNode = class(TParentNode)
  public
    constructor Create(Nodes: PNodeList);

    function Eval: float; override;

    {$IFDEF DUMP}
    function ToText: string; override;
    {$ENDIF}
  end;

{-- Term --}

function TAndNode.Eval: float;
begin
  if (FNodes[0].Eval = 0) or (FNodes[1].Eval = 0) then
    result := 0
  else
    result := 1;
end;

{$IFDEF DUMP}
function TAndNode.ToText: string;
begin
  Show('&');
end;
{$ENDIF}

{-- Sum --}

function TOrNode.Eval: float;
begin
  if (FNodes[0].Eval = 0) and (FNodes[1].Eval = 0) then
    result := 0
  else
    result := 1;
end;

{$IFDEF DUMP}
function TOrNode.ToText: string;
begin
  Show('|');
end;
{$ENDIF}


{-- Expression --}

function TLessNode.Eval: float;
begin
  if Nodes[0].Eval < Nodes[1].eval then
    result := 1.0
  else
    result := 0.0
end;

{$IFDEF DUMP}
function TLessNode.ToText: string;
begin
  result := Show('<');
end;
{$ENDIF}

function TLessOrEqualNode.Eval: float;
begin
  if Nodes[0].Eval <= Nodes[1].eval then
    result := 1.0
  else
    result := 0.0
end;

{$IFDEF DUMP}
function TLessOrEqualNode.ToText: string;
begin
  result := Show('<=');
end;
{$ENDIF}

function TGreaterNode.Eval: float;
begin
  if Nodes[0].Eval > Nodes[1].eval then
    result := 1.0
  else
    result := 0.0
end;

{$IFDEF DUMP}
function TGreaterNode.ToText: string;
begin
  result := Show('>');
end;
{$ENDIF}

function TGreaterOrEqualNode.Eval: float;
begin
  if Nodes[0].Eval >= Nodes[1].eval then
    result := 1.0
  else
    result := 0.0
end;

{$IFDEF DUMP}
function TGreaterOrEqualNode.ToText: string;
begin
  result := Show('>=');
end;
{$ENDIF}

function TEqualNode.Eval: float;
begin
  if Nodes[0].Eval = Nodes[1].eval then
    result := 1.0
  else
    result := 0.0
end;

{$IFDEF DUMP}
function TEqualNode.ToText: string;
begin
  result := Show('=');
end;
{$ENDIF}

function TNotEqualNode.Eval: float;
begin
  if Nodes[0].Eval <> Nodes[1].eval then
    result := 1.0
  else
    result := 0.0
end;

{$IFDEF DUMP}
function TNotEqualNode.ToText: string;
begin
  result := Show('<>');
end;
{$ENDIF}

constructor TIfNode.Create(Nodes: PNodeList);
begin
  FCount := 3;
  FNodes := Nodes;
end;

function TIfNode.Eval: float;
begin
  if Nodes[0].Eval = 0.0 then
    result := Nodes[2].Eval
  else
    result := Nodes[1].Eval
end;

{$IFDEF DUMP}
function TIfNode.ToText: string;
var
  I: Integer;
begin
  result := Format('%*.sif node%s', [FLevel,' ', inherited ToText]);
  incLevel;
  for I := 0 to FCount - 1 do
    result := result + #13#10 + FNodes[i].ToText;
  decLevel;
end;
{$ENDIF}

{-- factor --}

{ TNotNode }

function TNotNode.Eval: float;
begin
  if Nodes[0].eval = 0 then
    result := 1
  else
    result := 0
end;

{$IFDEF DUMP}
function TNotNode.ToText: string;
begin
  result := Format('%*.s! node%s)'#13#10, [FLevel,' ', inherited ToText]);
  incLevel;
  result := result + Nodes[0].ToText;
  decLevel;
end;
{$ENDIF}

{---------------------------------------------------}

Type
  TParamsList = class(TStringList)
  public
    UserFunction: TUserFunction;
    procedure Clear; override;
  end;

procedure TParamsList.Clear;
begin
  inherited Clear;
  UserFunction := nil;
end;

var
  Tokenizer: TMdTokenizer;
  ParamsList: TParamsList;
  GroupExponentsLeftToRight: boolean;

procedure RaiseParseException(Error: TParserError);
var
  E: EMathParserErr;
begin
  E := EMathParserErr.Create(_(SParseError));
  with E do begin
    TokenStart := Tokenizer.TokenStart;
    TokenLength := Tokenizer.TokenLength;
    ParserError := Error;
  end;
  Raise E;
end;

function StrToNumber(const value: string): float;
var
  p: integer;
  s: string;
begin
  {We will be lazy and use StrToFloat to convert
   the string to a number. However StrToFloat
   uses the locale's decimal symbol and
   furthermore StrToFloat('2E') returns 2E1 which
   is too ambiguous. So if the string ends in E
   an exception will be raised.}
  s := value;
  if upcase(s[length(s)]) <> 'E' then begin
    if ParserDecimalchar <> DecimalSeparator then begin
      p := pos(ParserDecimalChar, s);
      while p > 0 do begin
        s[p] := DecimalSeparator;
        p := pos(ParserDecimalChar, s);
      end;
    end;
    try
      result := StrToFloat(s);
      exit;
    except
      // fall through on E: EConvertError
    end;
  end;
  RaiseParseException(peInvalidNumber);
end;

function Factor: TTree; forward;

function Exponent: TTree;
begin
  result := Factor;
  try
    while Tokenizer.TokenType = ttRaise do begin
      Tokenizer.NextTrueToken;
      {$IFDEF EXPONENT_RIGHT_ASSOCIATIVE}
      result := TPowerNode.Create(result, Exponent)
      {$ELSE}
      result := TPowerNode.Create(result, Factor);
      {$ENDIF}
    end;
  except
    // end up here if an error occured parsing the recursed Exponent
    // need to release the memory allocated to the first factor
    result.Free;
    raise;
  end;
end;

function Term: TTree;
begin
  result := Exponent;
  try
    repeat
      if Tokenizer.TokenType = ttMult then begin
        Tokenizer.NextTrueToken;
        result := TMultNode.Create(result, Exponent);
      end
      else if Tokenizer.TokenType = ttDiv then begin
        Tokenizer.NextTrueToken;
        result := TDivNode.Create(result, Exponent);
      end
      else if Tokenizer.TokenType = ttColon then begin
        Tokenizer.NextTrueToken;
        result := TIntDivNode.Create(result, Exponent);
      end
      else if Tokenizer.TokenType = ttPercent then begin
        Tokenizer.NextTrueToken;
        result := TModNode.Create(result, Exponent);
      end
      else if Tokenizer.TokenType = ttAmpersand then begin
        Tokenizer.NextTrueToken;
        result := TAndNode.Create(result, Exponent);
      end
      else
        exit;
    until false;
  except
    // end up here if an error occured parsing the second factor
    // need to release the memory allocated to the first factor
    result.Free;
    raise;
  end;
end;

function Sum: TTree;
begin
  result := Term;
  try
    repeat
      if Tokenizer.TokenType = ttPlus then begin
        Tokenizer.NextTrueToken;
        result := TAddNode.Create(result, Term);
      end
      else if Tokenizer.TokenType = ttMinus then begin
        Tokenizer.NextTrueToken;
        result := TSubNode.Create(result, Term);
      end
      else if Tokenizer.TokenType = ttVertBar then begin
        Tokenizer.NextTrueToken;
        result := TOrNode.Create(result, Term);
      end
      else
        exit;
    until false;
  except
    // end up here if an error occured parsing the second term
    // need to release the memory allocated to the first term
    result.Free;
    raise;
  end;
end;


//Expression2 = Sum,  { ("<" | "<=" | "=" | "==" | '>' | '>=' | '<>' | '!='), Sum };
function Expression2: TTree;
begin
  result := Sum;
  try
    repeat
      if Tokenizer.TokenType = ttLess then begin
        Tokenizer.NextTrueToken;
        result := TLessNode.Create(result, Sum);
      end
      else if Tokenizer.TokenType = ttLessOrEqual then begin
        Tokenizer.NextTrueToken;
        result := TLessOrEqualNode.Create(result, Sum);
      end
      else if Tokenizer.TokenType = ttGreater then begin
        Tokenizer.NextTrueToken;
        result := TGreaterNode.Create(result, Sum);
      end
      else if Tokenizer.TokenType = ttGreaterOrEqual then begin
        Tokenizer.NextTrueToken;
        result := TGreaterOrEqualNode.Create(result, Sum);
      end
      else if Tokenizer.TokenType = ttEqual then begin
        Tokenizer.NextTrueToken;
        result := TEqualNode.Create(result, Sum);
      end
      else if Tokenizer.TokenType = ttNotEqual then begin
        Tokenizer.NextTrueToken;
        result := TNotEqualNode.Create(result, Sum);
      end
      else
        exit;
    until false;
  except
    // end up here if an error occured parsing the second term
    // need to release the memory allocated to the first term
    result.Free;
    raise;
  end;
end;

//Expression = Sum,  [ ("<" | "<=" | "=" | "==" | '>' | '>=' | '<>' | '!='), Sum ];
function Expression: TTree;
begin
  result := Sum;
  try
    if Tokenizer.TokenType = ttLess then begin
      Tokenizer.NextTrueToken;
      result := TLessNode.Create(result, Sum);
    end
    else if Tokenizer.TokenType = ttLessOrEqual then begin
      Tokenizer.NextTrueToken;
      result := TLessOrEqualNode.Create(result, Sum);
    end
    else if Tokenizer.TokenType = ttGreater then begin
      Tokenizer.NextTrueToken;
      result := TGreaterNode.Create(result, Sum);
    end
    else if Tokenizer.TokenType = ttGreaterOrEqual then begin
      Tokenizer.NextTrueToken;
      result := TGreaterOrEqualNode.Create(result, Sum);
    end
    else if Tokenizer.TokenType = ttEqual then begin
      Tokenizer.NextTrueToken;
      result := TEqualNode.Create(result, Sum);
    end
    else if Tokenizer.TokenType = ttNotEqual then begin
      Tokenizer.NextTrueToken;
      result := TNotEqualNode.Create(result, Sum);
    end
    else
      exit;
  except
    // end up here if an error occured parsing the second term
    // need to release the memory allocated to the first term
    result.Free;
    raise;
  end;
end;

function Params(var FoundNodes: pNodeList): integer;
var
  list: TList;
  i: integer;
  aNode: TTree;

  procedure FreeList;
  var
    i: integer;
  begin
    for I := 0 to List.Count - 1 do
      TTree(List[I]).free;
    list.Free;
    list := nil;
  end;

begin
  if Tokenizer.NextTrueToken <> ttLeftParenthesis then
    RaiseParseException(peExpectedLeftPar);
  Tokenizer.NextTrueToken;
  FoundNodes := nil;
  result := 0;
  list := TList.Create;
  try
    repeat
      aNode := Expression;
      list.Add(aNode);
      if Tokenizer.TokenType = ttRightParenthesis then begin
        //Tokenizer.NextTrueToken; // skip ')' wait, let Call do this
        result := list.Count;
        if result > 0 then begin
          getmem(FoundNodes, result*sizeof(TTree));
          for i := 0 to result - 1 do
            FoundNodes^[i] := TTree(List[i]);
        end;
        list.free;
        exit;
      end;
      if Tokenizer.TokenType <> ttSeparator then
        RaiseParseException(peExpectedSeparator);
      Tokenizer.NextTrueToken; // skip ','
    until false;
  except
    // end up here if an error occured parsing the list of arguments
    // need to release the memory allocated to the already parsed
    // trees and the memory allocated to the list of nodes.
    FreeList;
    freemem(FoundNodes);
    raise;
  end;
end;

function Call: TTree;
var
  Nodes: pNodeList;
  Target: TTarget;
  FoundCount: integer;
  n: integer;

  procedure FreeNodes;
  var
    i: integer;
  begin
    for I := 0 to FoundCount - 1 do
      Nodes^[i].Free;
    freemem(Nodes);
  end;

begin
  n := ParamsList.IndexOf(Tokenizer.Token);
  if n >= 0 then begin
    {$IFDEF DEBUG}
    Assert(assigned(ParamsList.UserFunction), 'Missing UserFunction in Call');
    {$ENDIF}
    result := TParameterNode.Create(ParamsList.UserFunction, n);
    Tokenizer.NextTrueToken;
    exit;
  end;

  if not ExternalTargets.Find(Tokenizer.Token, Target) then
    if not InternalTargets.Find(Tokenizer.Token, Target) then
       RaiseParseException(peUnknownIdentifier);

  {$IFDEF DEBUG}
  {if (Target is TVariable) and assigned(ParamsList.UserFunction) then
    found f(a, b) := 2*x where x is a variable - not allowed }
  Assert( not ((Target is TVariable) and (assigned(ParamsList.UserFunction))),
    'Mix up between formal and global variables');
  {$ENDIF}

  if Target = nil then begin
    FoundCount := Params(Nodes);
    if FoundCount <> 3 then begin
      FreeNodes;
      if FoundCount < Target.Count then
        RaiseParseException(peTooFewParams)
      else {if FoundCount > count then}
        RaiseParseException(peTooManyParams)
    end;
    result := TIfNode.Create(Nodes);
    Tokenizer.NextTrueToken; // skip ')'
    exit;
  end;

  if (Target.Count = 0) and not (Target is TVariableFunction) then begin
    Nodes := nil;
    Tokenizer.NextTrueToken;
  end
  else begin
    FoundCount := Params(Nodes);
    if Target is TVariableFunction then begin
      if FoundCount < TVariableFunction(Target).MinCount then begin
        FreeNodes;
        RaiseParseException(peTooFewParams);
      end;
      Target.Count := FoundCount // to set TCallNode's FCount
    end
    else begin
      {assert(Target is TFunction, 'oops not a TFunction in Call');}
      if FoundCount <> Target.Count then begin
        FreeNodes;
        if FoundCount < Target.Count then
          RaiseParseException(peTooFewParams)
        else {if FoundCount > count then}
          RaiseParseException(peTooManyParams)
      end;
    end;
    Tokenizer.NextTrueToken; // only skip ')' here so above errors better located
  end;
  result := TCallNode.Create(Target, Nodes);
end;

function Factor: TTree;
begin
  if Tokenizer.TokenType = ttNumber then begin
    result := TNumberNode.Create(StrToNumber(Tokenizer.Token));
    Tokenizer.NextTrueToken;
  end
  else if Tokenizer.TokenType = ttIdentifier then begin
    result := Call;
  end
  else if Tokenizer.TokenType = ttPlus then begin
    Tokenizer.NextTrueToken;  // skip and ignore leading '+'
    result := Factor;
  end
  else if Tokenizer.TokenType = ttMinus then begin
    Tokenizer.NextTrueToken;  // skip '-'
    result := TNegNode.Create(Factor);       // unary -
  end
  else if Tokenizer.TokenType = ttBang then begin
    Tokenizer.NextTrueToken;  // skip '!'
    result := TNotNode.Create(Factor);
  end
  else if Tokenizer.TokenType = ttLeftParenthesis then begin
    Tokenizer.NextTrueToken;  // skip '('
    result := Expression;
    if Tokenizer.TokenType <> ttRightParenthesis then begin
      // result is an allocated node that needs to be
      // freed now because of the parsing error
      result.Free;
      RaiseParseException(peExpectedRightPar);
    end;
    Tokenizer.NextTrueToken; // skip ')'
  end
  else if Tokenizer.TokenType = ttEOS then
    RaiseParseException(peUnexpectedEOS)
  else
    RaiseParseException(peUnexpectedElement);
end;

function Parse(const source: string): TTree;
begin
  tokenizer.Source := source;
  tokenizer.NextTrueToken; // prime the pump
  result := Expression;
  if (Tokenizer.TokenType <> ttEOS) then begin
    result.Free;
    RaiseParseException(peExpectedEOS);
  end;
end;


procedure ParseAssign(const source: string);
var
  Name: string;
  uf: TUserFunction;

  procedure ParseParamsList;
  begin
    while (Tokenizer.TokenType <> ttRightParenthesis) do begin
      if Tokenizer.NextTrueToken <> ttIdentifier then
        RaiseParseException(peExpectedFormalParameter)
      else if Tokenizer.Token = Name then // cannot have f(x, f) := 2
        RaiseParseException(peInvalidFormalParameterName)
      else if ParamsList.IndexOf(Tokenizer.Token) >= 0 then // cannot have f(x, x) := 2*x
        RaiseParseException(peDuplicateFormalParameterName);
      ParamsList.Add(Tokenizer.Token);
      if not (Tokenizer.NextTrueToken in [ttSeparator, ttRightParenthesis]) then
        RaiseParseException(peExpectedSeparatorOrRightPar);
    end;
  end;

begin
  tokenizer.Source := source;
  ParamsList.Clear;
  try
    // get the name
    if tokenizer.NextTrueToken <> ttIdentifier then
      RaiseParseException(peExpectedIdentifier);
    Name := tokenizer.Token;

    {$IFNDEF CAN_REDEFINE_BUILTIN_FUNCTION}
    if InternalTargets.HasTarget(Name) then
      RaiseParseException(peIdentIsBuiltInName);
    {$ENDIF}

    if ExternalTargets.HasTarget(Name) then
      RaiseParseException(peDuplicateIdentifier);

    // get param list if there's one
    if tokenizer.NextTrueToken = ttLeftParenthesis then begin
      ParseParamsList;
      if tokenizer.TokenType <> ttRightParenthesis then
        RaiseParseException(peExpectedRightPar);
      tokenizer.NextTrueToken;
    end;
    // find :=
    if tokenizer.TokenType <> ttAssign then
      RaiseParseException(peExpectedAssign);
    tokenizer.NextTrueToken;
    // get expression
    uf := TUserFunction.Create(nil, ParamsList.Count);
    ParamsList.UserFunction := uf;
    try
      uf.Tree := Expression;
      if Tokenizer.TokenType <> ttEOS then
        RaiseParseException(peExpectedEOS);
      ExternalTargets.Add(Name, uf);
    except
      uf.Free;
      raise;
    end;
  finally
    ParamsList.Clear;
  end;
end;

function EvaluateExpression(const source: string): float;
var
  tree: TTree;
begin
  tree := nil;
  try
    tree := Parse(source);
    result := tree.eval;
  finally
    tree.Free;
  end;
end;

function ParserDecimalChar: char;
begin
  result := tokenizer.DecimalChar;
end;

function SetParserDecimalChar(value: char): char;
begin
  result := tokenizer.DecimalChar;
  tokenizer.DecimalChar := value;
end;

function ParserListChar: char;
begin
  result := tokenizer.ListChar;
end;

function SetParserListChar(value: char): char;
begin
  result := tokenizer.ListChar;
  tokenizer.ListChar := value;
end;


initialization
  tokenizer := TMdTokenizer.create;
  ParamsList := TParamsList.Create;
finalization
  ParamsList.Free;
  tokenizer.free;
end.

