(***********************************************************************
 *
 * Nodes.inc version 1.0
 *
 * Copyright (c) 2014 Michel Deslierres
 * All rights reserved. Tous droits rservs.
 *
 * Read full license in the file license.txt
 * Lire la licence au complet dans le fichier license.txt
 *
 ************************************************************************
 *
 * Changements / Changes
 *
 * - version 1.0
 *      This file contains the definitions and implementation code of 
 *      classes used to represent nodes in a parse tree. The content 
 *      was taken from MathParser.pas version 1.0.
 *
 ************************************************************************)

type
  {Parse tree nodes often have children that are themselves
   node (TTree). These are held in an array allocated on
   the heap of type PNodeList}
  TNodeList = array[0..maxint div sizeof(TTree) - 1] of TTree;
  PNodeList = ^TNodeList;

  { Abstract parse tree nodes }

  TParentNode = class(TTree)
  private
    FCount: integer;
    FNodes: PNodeList;
    function Get(index: integer): TTree;
    procedure Put(index: integer; Node: TTree);
  public
    constructor Create(aCount: integer);
    destructor Destroy; override;
    property Nodes[index: integer]: TTree read Get write Put;
  end;

  TUnaryNode = class(TParentNode)
  public
    {Constructor of unary nodes that are nodes with a single child.
     The child node must be created first and passed as a parameter
     to this constructor.}
    constructor Create(aNode: TTree);
  end;

  TBinaryNode = class(TParentNode)
  private
    {$IFDEF DUMP}
    function Show(const value: string): string;
    {$ENDIF}
  public
    {Constructor of binary nodes that have two children. The child nodes must
    be created first and passed as parameters to this constructor.}
    constructor Create(aLeftNode, aRightNode: TTree);
  end;

  TCallNode = class(TParentNode)
  private
    FTarget: TTarget;
    FArgs: pFloatArray;
  public
    constructor Create(Target: TTarget; Nodes: PNodeList);
    destructor Destroy; override;
    function Eval: float; override;

    {$IFDEF DUMP}
    function ToText: string; override;
    {$ENDIF}
  end;

  { -- instantiated nodes -- }

  { Exponent node }

  TPowerNode = class(TBinaryNode)
  public
    function Eval: float; override;
    {$IFDEF DUMP}
    function ToText: string; override;
    {$ENDIF}
  end;

  { Term nodes }

  TMultNode = class(TBinaryNode)
  public
    function Eval: float; override;
    {$IFDEF DUMP}
    function ToText: string; override;
    {$ENDIF}
  end;

  TDivNode = class(TBinaryNode)
  public
    function Eval: float; override;
    {$IFDEF DUMP}
    function ToText: string; override;
    {$ENDIF}
  end;
  
  TIntDivNode = class(TBinaryNode)
  public
    function Eval: float; override;
    {$IFDEF DUMP}
    function ToText: string; override;
    {$ENDIF}
  end;
  
  TModNode = class(TBinaryNode)
  public
    function Eval: float; override;
    {$IFDEF DUMP}
    function ToText: string; override;
    {$ENDIF}
  end;

  { Expression nodes }

  TAddNode = class(TBinaryNode)
  public
    function Eval: float; override;
    {$IFDEF DUMP}
    function ToText: string; override;
    {$ENDIF}
  end;

  TSubNode = class(TBinaryNode)
  public
    function Eval: float; override;
    {$IFDEF DUMP}
    function ToText: string; override;
    {$ENDIF}
  end;

  { Factor nodes }

  TNumberNode = class(TTree)
  private
    FValue: float;
  public
    constructor Create(aValue: float);
    function Eval: float; override;
    {$IFDEF DUMP}
    function ToText: string; override;
    {$ENDIF}
  end;

  TNegNode = class(TUnaryNode)
  public
    function Eval: float; override;
    {$IFDEF DUMP}
    function ToText: string; override;
    {$ENDIF}
  end;

  TParameterNode = class(TTree)
  private
    FUserFunction: TUserFunction;
    FIndex: integer;
  public
    constructor Create(aUserFunction: TUserFunction; Index: integer);
    function Eval: float; override;
    {$IFDEF DUMP}
    function ToText: string; override;
    {$ENDIF}
  end;

{ -- abstract nodes -- }

constructor TParentNode.Create(aCount: integer);
var
   i: integer;
begin
  inherited create;
  FCount := aCount;
  if FCount = 0 then
    FNodes := nil
  else begin
    getmem(FNodes, FCount*sizeof(TTree));
    {$IFDEF DEBUG}
    for I := 0 to aCount - 1 do
      FNodes^[I] := nil;
    {$ENDIF}
  end;
end;

destructor TParentNode.Destroy;
var
   i: integer;
begin
  if FCount > 0 then begin
    for I := 0 to FCount - 1 do
      Nodes[I].Free;
    freemem(FNodes);
  end;
  inherited Destroy;
end;

function TParentNode.Get(index: integer): TTree;
begin
  {$IFDEF DEBUG}
  if (index < 0) or (index >= FCount) then
    raise ERangeError.CreateRes(@SRangeError);
  {$ENDIF}
  result := FNodes^[Index];
end;

procedure TParentNode.Put(index: integer; Node: TTree);
begin
  {$IFDEF DEBUG}
  if (index < 0) or (index >= FCount) then
    raise ERangeError.CreateRes(@SRangeError);
  {$ENDIF}
  FNodes^[Index] := Node;
end;

constructor TUnaryNode.Create(aNode: TTree);
begin
  inherited Create(1);
  Nodes[0] := aNode;
end;

constructor TBinaryNode.Create(aLeftNode, aRightNode: TTree);
begin
  inherited Create(2);
  Nodes[0] := aLeftNode;
  Nodes[1] := aRightNode;
end;

{$IFDEF DUMP}
function TBinaryNode.Show(const value: string): string;
begin
  result := Format('%*.s%s (%s) %s'+ NL, [FLevel,' ', ClassName, value, inherited ToText]);
  incLevel;
  result := result + Nodes[0].ToText;
  result := result + NL;
  result := result + Nodes[1].ToText;
  decLevel;
end;
{$ENDIF}

{ TPowerNode }

function TPowerNode.Eval: float;
begin
  result := Power(Nodes[0].eval, Nodes[1].eval)
end;

{$IFDEF DUMP}
function TPowerNode.ToText: string;
begin
  result := Show('^');
end;
{$ENDIF}

{-- Term --}

{ TMultNode }

function TMultNode.Eval: float;
begin
  result := Nodes[0].eval * Nodes[1].eval
end;

{$IFDEF DUMP}
function TMultNode.ToText: string;
begin
  result := Show('*');
end;
{$ENDIF}

{ TDivNode }

function TDivNode.Eval: float;
begin
  result := Nodes[0].eval / Nodes[1].eval
end;

{$IFDEF DUMP}
function TDivNode.ToText: string;
begin
  result := Show('/')
end;
{$ENDIF}

{ TIntDivNode }

function TIntDivNode.Eval: float;
begin
  {$IFDEF DELPHI_DIV_MOD}
  result := DivT(Nodes[0].Eval, Nodes[1].Eval);
  {$ELSE}
  result := DivE(Nodes[0].Eval, Nodes[1].Eval);
  {$ENDIF}
end;

{$IFDEF DUMP}
function TIntDivNode.ToText: string;
begin
  Show(':');
end;
{$ENDIF}

{ TModNode }

function TModNode.Eval: float;
begin
  {$IFDEF DELPHI_DIV_MOD}
  result := ModT(Nodes[0].Eval, Nodes[1].Eval);
  {$ELSE}
  result := ModE(Nodes[0].Eval, Nodes[1].Eval);
  {$ENDIF}
end;

{$IFDEF DUMP}
function TModNode.ToText: string;
begin
  Show('%')
end;
{$ENDIF}

{-- Expression --}

{ TAddNode }

function TAddNode.Eval: float;
begin
  result := Nodes[0].eval + Nodes[1].eval
end;

{$IFDEF DUMP}
function TAddNode.ToText: string;
begin
  result := Show('+');
end;
{$ENDIF}

{ TSubNode }

function TSubNode.Eval: float;
begin
  result := Nodes[0].eval - Nodes[1].eval
end;

{$IFDEF DUMP}
function TSubNode.ToText: string;
begin
  result := Show('-');
end;
{$ENDIF}

{-- Factor --}

constructor TNumberNode.Create(aValue: float);
begin
  inherited Create;
  FValue := avalue;
end;

function TNumberNode.Eval: float;
begin
  result := FValue;
end;

{$IFDEF DUMP}
function TNumberNode.ToText: string;
begin
  result := Format('%*.s%s%s', [FLevel,' ', ClassName, inherited ToText]);
end;
{$ENDIF}

{ TCallNode }

constructor TCallNode.Create(Target: TTarget; Nodes: PNodeList);
begin
  FTarget := Target;
  FNodes := Nodes;
  FCount := Target.Count;
  if FCount > 0 then begin
    getmem(FArgs, FCount*sizeof(float));
    {$IFDEF DEBUG}
    fillchar(FArgs^, FCount*sizeof(float), 0);
    {$ENDIF}
  end;
end;

destructor TCallNode.Destroy;
begin
  if FCount > 0 then
    freemem(FArgs);
  inherited;
end;

function TCallNode.Eval: float;
var
  I: Integer;
begin
  if FCount > 0 then begin
    for I := 0 to FCount - 1 do
      FArgs^[I] := Nodes[I].Eval;
    TFunction(FTarget).ArgArray := FArgs;
    if FTarget is TVariableFunction then
      TVariableFunction(FTarget).Count := FCount;
  end;
  result := FTarget.Calc;
end;

{$IFDEF DUMP}
function TCallNode.ToText: string;
var
  i: integer;
  targetName: string;
begin
  targetName := ExternalTargets.NameOf(FTarget);
  if (targetName = '') or (targetName[1] = '<') then
    targetName := InternalTargets.NameOf(FTarget);
  result := Format('%*.s%s (%s)%s', [FLevel,' ', ClassName,
    targetName, inherited ToText]);
  if FCount > 0 then begin
    incLevel;
    for I := 0 to FCount - 1 do
      result := result + #13#10 + FNodes[i].ToText;
    decLevel;
  end;
  if FTarget is TUserFunction then begin
    incLevel;
    result := Format('%s'#13#10'%*.sUserFunction %s:', [result, FLevel, ' ', targetName]);
    incLevel;
    result := result + #13#10 + TUserFunction(FTarget).Tree.ToText;
    decLevel;
    decLevel;
  end;
end;
{$ENDIF}

{ TNegNode }

function TNegNode.Eval: float;
begin
  result := - Nodes[0].eval
end;

{$IFDEF DUMP}
function TNegNode.ToText: string;
begin
  result := Format('%*.s%s (unary -)%s)'#13#10, [FLevel,' ', ClassName, inherited ToText]);
  incLevel;
  result := result + Nodes[0].ToText;
  decLevel;
end;
{$ENDIF}

{  TParameterNode }

constructor TParameterNode.Create(aUserFunction: TUserFunction; Index: integer);
begin
  inherited Create;
  FIndex := Index;
  FUserFunction := aUserFunction;
end;

function TParameterNode.Eval: float;
begin
  result := FUserFunction.Args[FIndex]
end;

{$IFDEF DUMP}
function TParameterNode.ToText: string;
begin
  result := Format('%*.sparameter %d node%s)', [FLevel,' ',FIndex, inherited ToText]);
end;
{$ENDIF}
