﻿unit TestExternalFunctions;

interface

{$I ..\Parser.inc}

uses
  TestFramework, Classes, SysUtils,
  math, jclMath,
  ParserTypes, ParserConsts;

type
  TTestMathCase = class(TTestCase)
  strict private
  protected
    function DoParse(const source: string): TTree;
    procedure DoParseAssign(const source: string);
    procedure TestOk(const value: string; wanted: float; eps: float = 0; const msg: string = '');
    procedure TestBadParse(const value: string; error: TParserError);
    procedure TestBadEval(const value: string; error: ExceptClass);
    procedure TestBadAssign(const value: string; error: TParserError);
  public
    procedure SetUp; override;
    procedure TearDown; override;
  end;

  TTestUserFunctions = class(TTestMathCase)
  published
    procedure TestBadParams;
    procedure TestUserFunctionCalc;
    procedure TestAssign;
    procedure TestMixedVariables;
  end;

implementation

uses
  Targets, MathParser;

const
  eps14 = 1E-14;
  eps13 = 1E-13;
  eps12 = 1E-12;
  eps11 = 1E-11;
  eps10 = 1E-10;
  eps9  = 1E-9;
  eps6  = 1E-6;


{ TTestMathCase }

function TTestMathCase.DoParse(const source: string): TTree;
begin
  try
    result := Parse(source);
  except
    On E: EMathParserErr do begin
      CheckEquals(ParseMessage(peNoError), ParseMessage(E.ParserError));
    end;
  end;
end;

procedure TTestMathCase.DoParseAssign(const source: string);
begin
  try
    ParseAssign(source);
  except
    On E: Exception do begin
      if E is EMathParserErr then with EMathParserErr(E) do begin
        CheckEquals(ParseMessage(peNoError), ParseMessage(ParserError));
      end
      else begin
        CheckTrue(false, E.Message);
      end;
    end;
  end;
end;

procedure TTestMathCase.TestOk(const value: string; wanted, eps: float; const msg: string);
var
  res: float;
begin
  if eps < 0 then begin
    res := EvaluateExpression(value);
    if abs((res-wanted)/wanted) > -eps then
      CheckTrue(false, Format('Expected %g got %g %s', [wanted, res, msg]))
    else
      CheckTrue(true, 'so DUnit does not complain about missing test');
  end
  else
    CheckEquals(wanted, EvaluateExpression(value), eps, msg);
end;

procedure TTestMathCase.SetUp;
begin
  inherited;
  SetParserDecimalChar('.');
  SetParserListChar(',');
  ExternalTargets.Clear;
end;

procedure TTestMathCase.TearDown;
begin
  inherited;
  ExternalTargets.Clear;
end;

procedure TTestMathCase.TestBadEval(const value: string; error: ExceptClass);
begin
  try
    EvaluateExpression(value);
  except
    On E: Exception do begin
      CheckEquals(error.ClassName, E.ClassName);
    end;
  end;
end;

procedure TTestMathCase.TestBadParse(const value: string; error: TParserError);
begin
  try
    EvaluateExpression(value);
  except
    On E: EMathParserErr do begin
      CheckEquals(ord(error), ord(E.ParserError));
    end;
  end;
end;

procedure TTestMathCase.TestBadAssign(const value: string; error: TParserError);
begin
  try
    ParseAssign(value);
  except
    On E: EMathParserErr do begin
      CheckEquals(ord(error), ord(E.ParserError));
    end;
  end;
end;

procedure TTestUserFunctions.TestBadParams;
begin
  TestBadParse('lin(32)', peUnknownIdentifier);
  TestBadAssign('3pi(x) := 8*x', peExpectedIdentifier);
  TestBadAssign('f(x) = 2*x', peExpectedAssign);
  {$IFDEF CAN_REDEFINE_BUILTIN_FUNCTION}
  DoParseAssign('sin(x) := x*sin(x)');
  TestOk('sin(5)', 5*sin(5), 0, 'redefined sin');
  {$ELSE}
  TestBadAssign('sin(y) := cos(y)', peIdentIsBuiltInName);
  {$ENDIF}
  ParseAssign('f(x) := 2*x');
  CheckTrue(ExternalTargets.HasTarget('f'), 'f not added');
  TestBadAssign('f(a, b) := a/b', peDuplicateIdentifier);
  ExternalTargets.Clear;
  TestBadAssign('f() := 5', peExpectedFormalParameter);
  TestBadAssign('f(a, ) := 5*a + b', peExpectedFormalParameter);
  TestBadAssign('f(a, 32) := 5', peExpectedFormalParameter);
  TestBadAssign('f(a, f) := a*f', peInvalidFormalParameterName);
  TestBadAssign('f(x, a, x, b) := 5*x', peDuplicateFormalParameterName);
  TestBadAssign('f(x 32 y) := x/y', peExpectedSeparatorOrRightPar);
  end;

procedure TTestUserFunctions.TestAssign;
begin
  CheckFalse(ExternalTargets.HasTarget('f'), 'Can''t do test "f" already added');
  ParseAssign('f(x, y) := 2*x + y');
  CheckTrue(ExternalTargets.HasTarget('f'), 'Can''t find "f"');
  testOk('f(4, 3)', 11);
  testOk('f(sin(4), 3)', 2*sin(4) + 3);
  ExternalTargets.Delete('f');
  CheckFalse(ExternalTargets.HasTarget('f'), '"f" found after deletion');
end;

procedure TTestUserFunctions.TestUserFunctionCalc;
var
  aTree: TTree;
begin
  CheckFalse(ExternalTargets.HasTarget('user'), 'Can''t do test "user" already added');
  aTree := Parse('32');
  ExternalTargets.Add('user', TUserFunction.Create(aTree, 0));
  TestOk('user', 32);
  TestOk('2*user', 64);
  ExternalTargets.Clear;
  CheckFalse(ExternalTargets.HasTarget('user'), '"user" found after clear');
end;

procedure TTestUserFunctions.TestMixedVariables;
var
  x: float;
  msg: string;
  res: boolean;
begin
  ExternalTargets.Add('x', TVariable.create(@x));
  try
    ParseAssign('f(a, b) := 2*x + a/b');
  except
    On E: Exception do begin
      res := E is EAssertionFailed;
      if not res then
        msg := E.ClassName + ' ' + E.Message;
    end;
  end;
  CheckTrue(res, 'Expected EAssertionFailed got '+ msg);
end;


initialization
  // Recenser tous les cas de test avec le lanceur de test
  RegisterTest(TTestUserFunctions.Suite);
end.

