﻿unit TestTargets;


interface

{$I ..\Parser.inc}

uses
  TestFramework, Classes, SysUtils,  math,
  ParserTypes;

type
  TTestMathCase = class(TTestCase)
  strict private
  protected
    procedure TestOk(const value: string; wanted: float; eps: float = 0; const msg: string = '');
    procedure TestBadParse(const value: string; error: TParserError);
    procedure TestBadEval(const value: string; error: ExceptClass);
  public
    procedure SetUp; override;
  end;

  TestConstants = class(TTestMathCase)
  published
    procedure Teste;
    procedure TestPi;
    procedure TestCombo;
  end;

  TestMathFunctions = class(TTestMathCase)
  published
    procedure TestAbs;
    procedure TestOdd;
    procedure TestFloor;
    procedure TestCeiling;
    procedure TestInt;
    procedure TestSgn;
    procedure TestRound;
    procedure TestClamp;
    procedure TestFactorial;
    procedure TestIns;
    procedure TestOuts;
    procedure TestNear;
  end;

  TestPowerFunctions = class(TTestMathCase)
  published
    procedure TestExp;
    procedure TestLn;
    procedure TestLog;
    procedure TestLog10;
  end;

  TestTrigonometricFunctions = class(TTestMathCase)
  published
    procedure TestSin;
    procedure TestCos;
    procedure TestTan;
    procedure TestCot;
    procedure TestSec;
    procedure TestCsc;
  end;

  TestInverseTrigonometricFunctions = class(TTestMathCase)
  published
    procedure TestASin;
    procedure TestACos;
    procedure TestATan;
    procedure TestACot;
    procedure TestASec;
    procedure TestACsc;
  end;

  TestParserAngle = class(TTestMathCase)
  strict private
    d: array[0..24] of float;
    r: array[0..24] of float;
  public
    procedure SetUp; override;
  published
    procedure TestRad;
    procedure TestDeg;
  end;

  TTestHyperbolic = class(TTestMathCase)
  published
    procedure TestSinh;
    procedure TestCosh;
    procedure TestTanh;
    procedure TestCoth;
    procedure TestCsch;
    procedure TestSech;
  end;

  TTestInverseHyperbolic = class(TTestMathCase)
  published
    procedure TestASinh;
    procedure TestACosh;
    procedure TestATanh;
    procedure TestACoth;
    procedure TestACsch;
    procedure TestASech;
  end;

  TTestVariableFunctions = class(TTestMathCase)
  published
    procedure TestAvg;
    procedure TestMin;
    procedure TestMax;
    procedure TestSsq;
    procedure TestVar;
    procedure TestVarp;
    procedure TestPoly;
    procedure TestRecursion;
  end;

implementation

uses
  ParserConsts, Targets, MathParser;

const
  eps14 = 1E-14;
  eps13 = 1E-13;
  eps12 = 1E-12;
  eps11 = 1E-11;
  eps10 = 1E-10;
  eps9  = 1E-9;
  eps6  = 1E-6;

{ TTestMathCase }

procedure TTestMathCase.SetUp;
begin
  inherited;
  SetParserDecimalChar('.');
  SetParserListChar(',');
end;

procedure TTestMathCase.TestOk(const value: string; wanted, eps: float; const msg: string);
var
  res: float;
begin
  if eps < 0 then begin
    res := EvaluateExpression(value);
    if abs((res-wanted)/wanted) > -eps then
      CheckTrue(false, Format('Expected %g got %g %s', [wanted, res, msg]))
    else
      CheckTrue(true, 'so DUnit does not complain about missing test');
  end
  else
    CheckEquals(wanted, EvaluateExpression(value), eps, msg);
end;

procedure TTestMathCase.TestBadEval(const value: string; error: ExceptClass);
begin
  try
    EvaluateExpression(value);
  except
    On E: Exception do begin
      CheckEquals(error.ClassName, E.ClassName);
    end;
  end;
end;

procedure TTestMathCase.TestBadParse(const value: string; error: TParserError);
begin
  try
    EvaluateExpression(value);
  except
    On E: EMathParserErr do begin
      CheckEquals(ParseMessage(error), ParseMessage(E.ParserError));
    end;
  end;
end;

{ TTestMathFunctions }

procedure TestMathFunctions.TestAbs;
begin
  TestOk('abs(0)', abs(0));
  TestOk('abs(1)', abs(1));
  TestOk('abs(-1)', abs(-1));
end;


procedure TestMathFunctions.TestCeiling;
begin
  TestOk('ceil(3.8)', 4);
end;

procedure TestMathFunctions.TestClamp;
begin
  TestOk('clamp(5, 2, 8)', 5);
  TestOk('clamp(-5, 2, 8)', 2);
  TestOk('clamp(15, 2, 8)', 8);
end;

procedure TestMathFunctions.TestFactorial;
begin
  TestOk('fact(4)', 24);
  //TestOk('fact(-2)', 0);
  TestBadEval('fact(4000)', EOverflow);
end;

procedure TestMathFunctions.TestFloor;
begin
  TestOk('floor(3.8)', 3);
end;

procedure TestMathFunctions.TestIns;
begin
  TestOk('ins(5, 2, 8)', 1);
  TestOk('ins(-5, 2, 8)', 0);
  TestOk('ins(15, 2, 8)', 0);
end;

procedure TestMathFunctions.TestInt;
begin
  TestOk('int(5.7)', 5);
  TestOk('int(-5.7)', -5);
end;

procedure TestMathFunctions.TestNear;
begin
  TestOk('near(1, 1 + 1E-18, 0)', 0);
  TestOk('near(1, 1 + 1E-19, 0)', 1);

  TestOk('near(1E18, 1E18 + 1, 0)', 0);
  TestOk('near(1E19, 1E19 + 1, 0)', 1);
end;

procedure TestMathFunctions.TestOdd;
begin
  TestOk('odd(3)', 1);
  TestOk('odd(4)', 0);
end;

procedure TestMathFunctions.TestOuts;
begin
  TestOk('outs(5, 2, 8)', 0);
  TestOk('outs(-5, 2, 8)', 1);
  TestOk('outs(15, 2, 8)', 1);
end;

procedure TestMathFunctions.TestRound;
begin
  TestOk('round(5.3)', 5);
  TestOk('round(5.7)', 6);
  TestOk('round(-5.3)', -5);
  //TestOk('round(-5.7)', round(-5.7));
end;

procedure TestMathFunctions.TestSgn;
begin
  TestOk('sgn(0)', 0);
  TestOk('sgn(-0.2)', -1);
  TestOk('sgn(20)', 1);
end;

{ TestPowerFunctions }

procedure TestPowerFunctions.TestExp;
begin
  TestOk('exp(3.234)', exp(3.234));
  TestOk('exp(2.1)', exp(2.1));
  TestOk('exp(1.0)', exp(1.0));
  TestOk('exp(0.0)', exp(0.0));
  TestOk('exp(-1.0)', exp(-1.0));
  TestOk('exp(-2.1)', exp(-2.1));
  TestOk('exp(-3.234)', exp(-3.234));
end;

procedure TestPowerFunctions.TestLn;
begin
  TestOk('ln(3.234)', system.ln(3.234));
  TestOk('ln(2.1)', system.ln(2.1));
  TestOk('ln(1.0)', system.ln(1.0));
end;

procedure TestPowerFunctions.TestLog;
begin
  {$IFDEF USE_MATH_MINI}
  TestOk('log(5.01, 3.234)', logn(3.234, 5.01));
  TestOk('log(5.02, 2.1)', logn(2.1, 5.02));
  {$ELSE}
  TestOk('log(5.01, 3.234)', logn(3.234, 5.01), -1E-16);
  TestOk('log(5.02, 2.1)', logn(2.1, 5.02), -1E-16);
  {$ENDIF}
end;

procedure TestPowerFunctions.TestLog10;
var
  L: array[-6..6] of extended;
  I: integer;
  s: string;
begin
  L[0] := 1;
  for I := 1 to 6 do
    L[I] := 10*L[I-1];
  for I := -1 downto -6 do
    L[I] := L[I+1]/10;
  for I := -6 to 6 do begin
    str(L[I], s);
    TestOk(Format('log10(%s)', [s]), I, 1E-12, 'index: '+inttostr(I));
  end;
end;

function NForm(value: extended): string;
begin
  str(value, result);
end;

{ TestTrigonometricFunctions }

{
sin(A) = y/r  sin(A)
cos(A) = x/r  cos(A)
tan(A) = y/x  tan(A) = sin(A)/cos(A)
cot(A) = x/y  cot(A) = cos(A)/sin(A)
sec(A) = r/x  sec(A) = 1/cos(A)
csc(A) = r/y  csc(A) = 1/sin(A)
}

procedure TestTrigonometricFunctions.TestCos;

const
  count = 32;
var
  i: integer;
  a: float;
  step: float;
begin
  step := (2*pi)/count;
  a := 0;
  for I := 0 to Count - 1 do begin
    TestOk(Format('cos(%s)', [NForm(a)]), system.cos(a), eps14);
    a := a + step;
  end;
end;

procedure TestTrigonometricFunctions.TestCot;
const
  count = 32;
var
  i: integer;
  a: float;
  step: float;
  s,c: float;
begin
  step := (2*pi)/count;
  a := 0;
  for I := 0 to Count - 1 do begin
    c := cos(a);
    s := sin(a);
    if not isZero(s) then
      TestOk(Format('cot(%s)', [NForm(a)]), c/s, eps12, Format('angle: %g, index: %d', [a, i]));
    a := a + step;
  end;
end;

procedure TestTrigonometricFunctions.TestCsc;
const
  count = 32;
var
  i: integer;
  a: float;
  step: float;
  s: float;
begin
  step := (2*pi)/count;
  a := 0;
  for I := 0 to Count - 1 do begin
    s := sin(a);
    if not isZero(s) then
      TestOk(Format('csc(%s)', [NForm(a)]), 1/s, eps12, Format('angle: %g, index: %d', [a, i]));
    a := a + step;
  end;
end;

procedure TestTrigonometricFunctions.TestSec;
const
  count = 32;
var
  i: integer;
  a: float;
  step: float;
  c: float;
begin
  step := (2*pi)/count;
  a := 0;
  for I := 0 to Count - 1 do begin
    c := cos(a);
    if not isZero(c) then
      TestOk(Format('sec(%s)', [NForm(a)]), 1/c, eps12, Format('angle: %g, index: %d', [a, i]));
    a := a + step;
  end;
end;

procedure TestTrigonometricFunctions.TestSin;
const
  count = 32;
var
  i: integer;
  a: float;
  step: float;
begin
  step := (2*pi)/count;
  a := 0;
  for I := 0 to Count - 1 do begin
    TestOk(Format('sin(%s)', [NForm(a)]), system.sin(a), eps14);
    a := a + step;
  end;
end;

procedure TestTrigonometricFunctions.TestTan;
const
  count = 32;
var
  i: integer;
  a: float;
  step: float;
  s,c: float;
begin
  step := (2*pi)/count;
  a := 0;
  for I := 0 to Count - 1 do begin
    c := cos(a);
    s := sin(a);
    if not isZero(c) then
      TestOk(Format('tan(%s)', [NForm(a)]), s/c, eps12, Format('angle: %g, index: %d', [a, i]));
    a := a + step;
  end;
end;

{ TestInverseTrigonometricFunctions }
(*
asin(x) = A ⇔ sin(A) = x et A ∈ [-π/2..π/2]
acos(x) = A ⇔ cos(A) = x et A ∈ [0..π]
atan(x) = A ⇔ tan(A) = x et A ∈ ]-π/2..π/2[
acot(x) = A ⇔ cot(A) = x et A ∈ ]-π/2..π/2[
asec(x) = A ⇔ sec(A) = x et A ∈ [0..π/2]
acsc(x) = A ⇔ csc(A) = x et A ∈ [0..π/2]
*)

procedure TestInverseTrigonometricFunctions.TestACos;
// acos(x) = A ⇔cos(A) = x et A ∈[0..π]
const
  count = 32;
var
  min, max: float;
  step: float;
  i: integer;
  a, cosa: float;
begin
  max := pi;
  min := 0;
  step := (max-min)/count;
  a := min;
  for I := 0 to Count - 1 do begin
    cosa := system.Cos(a);
    TestOk(Format('acos(%s)', [NForm(cosa)]), a, eps14, Format('index: %d', [i]));
    a := a + step;
  end;
end;

procedure TestInverseTrigonometricFunctions.TestACot;
// acot(x) = A ⇔ cot(A) = x et A ∈ ]-π/2..π/2[
const
  count = 32;
var
  min, max: float;
  i: integer;
  a: float;
  step: float;
  s,c: float;
begin
  max := pi/2;
  min := -pi/2;
  step := (max-min)/count;
  a := min + step;
  for I := 1 to Count - 2 do begin
    c := cos(a);
    s := sin(a);
    if not isZero(s) then
      TestOk(Format('acot(%s)', [NForm(c/s)]), a, eps12, Format('angle: %g, index: %d', [a, i]));
    a := a + step;
  end;
end;

procedure TestInverseTrigonometricFunctions.TestACsc;
// acsc(x) = A ⇔csc(A) = x et A ∈[0..π/2]
// csc(A) = r/y = 1/sin(A)
const
  count = 32;
var
  min, max: float;
  i: integer;
  a: float;
  step: float;
  s: float;
begin
  max := pi/2;
  min := 0;
  step := (max-min)/count;
  a := min;
  for I := 0 to Count - 1 do begin
    s := sin(a);
    if not isZero(s) then
      TestOk(Format('acsc(%s)', [NForm(1/s)]), a, eps12, Format('index: %d', [i]));
    a := a + step;
  end;
end;

procedure TestInverseTrigonometricFunctions.TestASec;
// acsc(x) = A ⇔csc(A) = x et A ∈[0..π/2]
const
  count = 32;
var
  min, max: float;
  i: integer;
  a: float;
  step: float;
  c: float;
begin
  max := pi/2;
  min := 0;
  step := (max-min)/count;
  a := min;
  for I := 0 to Count - 1 do begin
    c := cos(a);
    if not isZero(c) then
      TestOk(Format('asec(%s)', [NForm(1/c)]), a, eps12, Format('angle: %g, index: %d', [a, i]));
    a := a + step;
  end;
end;

procedure TestInverseTrigonometricFunctions.TestASin;
// asin(x) = A ⇔sin(A) = x et A ∈[-π/2..π/2]
const
  count = 32;
var
  min, max: float;
  i: integer;
  a: float;
  step: float;
begin
  max := pi/2;
  min := -pi/2;
  step := (max-min)/count;
  a := min;
  for I := 0 to Count - 1 do begin
    TestOk(Format('asin(%s)', [NForm(system.sin(a))]), a, eps14);
    a := a + step;
  end;
end;

procedure TestInverseTrigonometricFunctions.TestATan;
// atan(x) = A ⇔ tan(A) = x et A ∈ ]-π/2..π/2[
const
  count = 32;
var
  min, max: float;
  i: integer;
  a: float;
  step: float;
  s,c: float;
begin
  max := -pi/2;
  min := pi/2;
  step := (max-min)/count;
  a := min + step;
  for I := 1 to Count - 2 do begin
    c := cos(a);
    s := sin(a);
    if not isZero(c) then
      TestOk(Format('atan(%s)', [NForm(s/c)]), a, eps12, Format('index: %d', [i]));
    a := a + step;
  end;
end;


{ TestParserAngle }

procedure TestParserAngle.SetUp;
var
  i: Integer;
  dd, rr: float;
  piby12: extended;
begin
  inherited;
  dd := 0;
  rr := 0;
  piBy12 := pi/12;
  for i := 0 to 24 do begin
    d[i] := dd;
    r[i] := rr;
    dd := dd + 15;
    rr := rr + piBy12;
  end;
end;

procedure TestParserAngle.TestDeg;
var
  i: integer;
begin
  for I := 0 to 24 do
    TestOk(Format('deg(%s)', [NForm(r[i])]), d[i], eps12, Format('Index %d', [I]));
end;

procedure TestParserAngle.TestRad;
var
  i: integer;
begin
  for I := 0 to 24 do
    TestOk(Format('rad(%s)', [NForm(d[i])]), r[i], eps12, Format('Index %d', [I]));
end;

type
  Rec = record
    angle, val : extended;
  end;

const
  SinhValues: array[0..23] of rec = (
    (angle:	-6.283185307179590	; val:	-267.744894041016000	),
    (angle:	-5.759586531581290	; val:	-158.606995056971000	),
    (angle:	-5.235987755982990	; val:	-93.954653468459900	),
    (angle:	-4.712388980384690	; val:	-55.654397599417500	),
    (angle:	-4.188790204786390	; val:	-32.963900290099900	),
    (angle:	-3.665191429188090	; val:	-19.519007046411200	),
    (angle:	-3.141592653589790	; val:	-11.548739357257700	),
    (angle:	-2.617993877991490	; val:	-6.817623304126540	),
    (angle:	-2.094395102393200	; val:	-3.998691342799820	),
    (angle:	-1.570796326794900	; val:	-2.301298902307290	),
    (angle:	-1.047197551196600	; val:	-1.249367050523980	),
    (angle:	-0.523598775598299	; val:	-0.547853473888040	),
    (angle:	0.000000000000000	; val:	0.000000000000000	),
    (angle:	0.523598775598298	; val:	0.547853473888039	),
    (angle:	1.047197551196600	; val:	1.249367050523970	),
    (angle:	1.570796326794900	; val:	2.301298902307290	),
    (angle:	2.094395102393190	; val:	3.998691342799820	),
    (angle:	2.617993877991490	; val:	6.817623304126540	),
    (angle:	3.141592653589790	; val:	11.548739357257700	),
    (angle:	3.665191429188090	; val:	19.519007046411100	),
    (angle:	4.188790204786390	; val:	32.963900290099900	),
    (angle:	4.712388980384690	; val:	55.654397599417500	),
    (angle:	5.235987755982990	; val:	93.954653468459800	),
    (angle:	5.759586531581290	; val:	158.606995056971000	));

procedure TTestHyperbolic.TestSinh;
var
  i: integer;
begin
  for i := 0 to 23 do
    TestOk(Format('sinh(%s)', [NForm(SinhValues[i].angle)]), SinhValues[i].val, eps11);
end;

procedure TTestInverseHyperbolic.TestASinh;
var
  i: integer;
begin
  for i := 0 to 23 do
    TestOk(Format('asinh(%s)', [NForm(SinhValues[i].val)]), SinhValues[i].angle, eps11);
end;

const
  CoshValues: array[0..23] of rec = (
    (angle:	-6.283185307179590	; val:	267.746761483748000	),
    (angle:	-5.759586531581290	; val:	158.610147471724000	),
    (angle:	-5.235987755982990	; val:	93.959975033938700	),
    (angle:	-4.712388980384690	; val:	55.663380890438700	),
    (angle:	-4.188790204786390	; val:	32.979064909964500	),
    (angle:	-3.665191429188090	; val:	19.544606316778300	),
    (angle:	-3.141592653589790	; val:	11.591953275521500	),
    (angle:	-2.617993877991490	; val:	6.890572364975880	),
    (angle:	-2.094395102393200	; val:	4.121836053869960	),
    (angle:	-1.570796326794900	; val:	2.509178478658060	),
    (angle:	-1.047197551196600	; val:	1.600286857702390	),
    (angle:	-0.523598775598299	; val:	1.140238321076430	),
    (angle:	0.000000000000000	; val:	1.000000000000000	),
    (angle:	0.523598775598298	; val:	1.140238321076430	),
    (angle:	1.047197551196600	; val:	1.600286857702390	),
    (angle:	1.570796326794900	; val:	2.509178478658050	),
    (angle:	2.094395102393190	; val:	4.121836053869950	),
    (angle:	2.617993877991490	; val:	6.890572364975870	),
    (angle:	3.141592653589790	; val:	11.591953275521500	),
    (angle:	3.665191429188090	; val:	19.544606316778200	),
    (angle:	4.188790204786390	; val:	32.979064909964400	),
    (angle:	4.712388980384690	; val:	55.663380890438700	),
    (angle:	5.235987755982990	; val:	93.959975033938600	),
    (angle:	5.759586531581290	; val:	158.610147471724000	));

procedure TTestHyperbolic.TestCosh;
var
  i: integer;
begin
  for i := 0 to 23 do
    TestOk(Format('cosh(%s)', [NForm(CoshValues[i].angle)]), CoshValues[i].val, eps11, Format('Index %d', [I]));
end;

procedure TTestInverseHyperbolic.TestACosh;
var
  i: integer;
begin
  // defined x >= 1 only
  for i := 12 to 23 do
    TestOk(Format('acosh(%s)', [NForm(CoshValues[i].val)]), CoshValues[i].angle, eps11, Format('Index %d', [I]));
end;

const
  TanhValues: array[0..23] of rec = (
    (angle:	-6.283185307179590	; val:	-0.999993025339611	),
    (angle:	-5.759586531581290	; val:	-0.999980124759965	),
    (angle:	-5.235987755982990	; val:	-0.999943363485603	),
    (angle:	-4.712388980384690	; val:	-0.999838613988633	),
    (angle:	-4.188790204786390	; val:	-0.999540174352852	),
    (angle:	-3.665191429188090	; val:	-0.998690213046393	),
    (angle:	-3.141592653589790	; val:	-0.996272076220750	),
    (angle:	-2.617993877991490	; val:	-0.989413207352682	),
    (angle:	-2.094395102393200	; val:	-0.970123821165931	),
    (angle:	-1.570796326794900	; val:	-0.917152335667274	),
    (angle:	-1.047197551196600	; val:	-0.780714435359268	),
    (angle:	-0.523598775598299	; val:	-0.480472778156452	),
    (angle:	0.000000000000000	; val:	0.000000000000000	),
    (angle:	0.523598775598298	; val:	0.480472778156451	),
    (angle:	1.047197551196600	; val:	0.780714435359268	),
    (angle:	1.570796326794900	; val:	0.917152335667274	),
    (angle:	2.094395102393190	; val:	0.970123821165931	),
    (angle:	2.617993877991490	; val:	0.989413207352682	),
    (angle:	3.141592653589790	; val:	0.996272076220750	),
    (angle:	3.665191429188090	; val:	0.998690213046393	),
    (angle:	4.188790204786390	; val:	0.999540174352852	),
    (angle:	4.712388980384690	; val:	0.999838613988633	),
    (angle:	5.235987755982990	; val:	0.999943363485603	),
    (angle:	5.759586531581290	; val:	0.999980124759965	));

procedure TTestHyperbolic.TestTanh;
var
  i: integer;
begin
  for i := 0 to 23 do
    TestOk(Format('tanh(%s)', [NForm(TanhValues[i].angle)]), TanhValues[i].val, eps11);
end;

procedure TTestInverseHyperbolic.TestATanh;
var
  i: integer;
begin
  for i := 0 to 23 do
    TestOk(Format('atanh(%s)', [NForm(TanhValues[i].val)]), TanhValues[i].angle, eps10);
end;

const
  CothValues: array[0..22] of rec = (
    (angle:	-6.283185307179590	; val:	-1.000006974709040	),
    (angle:	-5.759586531581290	; val:	-1.000019875635070	),
    (angle:	-5.235987755982990	; val:	-1.000056639722270	),
    (angle:	-4.712388980384690	; val:	-1.000161412061020	),
    (angle:	-4.188790204786390	; val:	-1.000460037184040	),
    (angle:	-3.665191429188090	; val:	-1.001311504745410	),
    (angle:	-3.141592653589790	; val:	-1.003741873197320	),
    (angle:	-2.617993877991490	; val:	-1.010700072091870	),
    (angle:	-2.094395102393200	; val:	-1.030796253202160	),
    (angle:	-1.570796326794900	; val:	-1.090331410727370	),
    (angle:	-1.047197551196600	; val:	-1.280878071045040	),
    (angle:	-0.523598775598299	; val:	-2.081283363933640	),
    //(angle:	0.000000000000000	; val:	#NUM !	),
    (angle:	0.523598775598298	; val:	2.081283363933640	),
    (angle:	1.047197551196600	; val:	1.280878071045050	),
    (angle:	1.570796326794900	; val:	1.090331410727370	),
    (angle:	2.094395102393190	; val:	1.030796253202160	),
    (angle:	2.617993877991490	; val:	1.010700072091870	),
    (angle:	3.141592653589790	; val:	1.003741873197320	),
    (angle:	3.665191429188090	; val:	1.001311504745410	),
    (angle:	4.188790204786390	; val:	1.000460037184040	),
    (angle:	4.712388980384690	; val:	1.000161412061020	),
    (angle:	5.235987755982990	; val:	1.000056639722270	),
    (angle:	5.759586531581290	; val:	1.000019875635070	));

procedure TTestHyperbolic.TestCoth;
var
  i: integer;
begin
  for i := 0 to 22 do
    TestOk(Format('coth(%s)', [NForm(CothValues[i].angle)]), CothValues[i].val, eps11);
end;

procedure TTestInverseHyperbolic.TestACoth;
var
  i: integer;
begin
  for i := 0 to 22 do
    TestOk(Format('acoth(%s)', [NForm(CothValues[i].val)]), CothValues[i].angle, eps9);
end;

const
  CschValues: array[0..22] of rec = (
    (angle:	-6.283185307179590	; val:	-0.003734898488286	),
    (angle:	-5.759586531581290	; val:	-0.006304892162170	),
    (angle:	-5.235987755982990	; val:	-0.010643432369592	),
    (angle:	-4.712388980384690	; val:	-0.017968032053777	),
    (angle:	-4.188790204786390	; val:	-0.030336216018113	),
    (angle:	-3.665191429188090	; val:	-0.051232114298758	),
    (angle:	-3.141592653589790	; val:	-0.086589537530047	),
    (angle:	-2.617993877991490	; val:	-0.146678681908800	),
    (angle:	-2.094395102393200	; val:	-0.250081817842888	),
    (angle:	-1.570796326794900	; val:	-0.434537208094696	),
    (angle:	-1.047197551196600	; val:	-0.800405292888592	),
    (angle:	-0.523598775598299	; val:	-1.825305574687950	),
    //(angle:	0.000000000000000	; val:	#DIV/0 !	),
    (angle:	0.523598775598298	; val:	1.825305574687960	),
    (angle:	1.047197551196600	; val:	0.800405292888593	),
    (angle:	1.570796326794900	; val:	0.434537208094696	),
    (angle:	2.094395102393190	; val:	0.250081817842889	),
    (angle:	2.617993877991490	; val:	0.146678681908800	),
    (angle:	3.141592653589790	; val:	0.086589537530047	),
    (angle:	3.665191429188090	; val:	0.051232114298758	),
    (angle:	4.188790204786390	; val:	0.030336216018113	),
    (angle:	4.712388980384690	; val:	0.017968032053777	),
    (angle:	5.235987755982990	; val:	0.010643432369592	),
    (angle:	5.759586531581290	; val:	0.006304892162170	));

procedure TTestHyperbolic.TestCsch;
var
  i: integer;
begin
  for i := 0 to 22 do
    TestOk(Format('csch(%s)', [NForm(CschValues[i].angle)]), CschValues[i].val, eps11);
end;

procedure TTestInverseHyperbolic.TestACsch;
var
  i: integer;
begin
  for i := 3 to 22 do
    TestOk(Format('acsch(%s)', [NForm(CschValues[i].val)]), CschValues[i].angle, eps11, Format('Index %d', [I]));
end;

const
  SechValues: array[0..23] of rec = (
    (angle:	-6.283185307179590	; val:	0.003734872438637	),
    (angle:	-5.759586531581290	; val:	0.006304766850925	),
    (angle:	-5.235987755982990	; val:	0.010642829562681	),
    (angle:	-4.712388980384690	; val:	0.017965132264752	),
    (angle:	-4.188790204786390	; val:	0.030322266647950	),
    (angle:	-3.665191429188090	; val:	0.051165011143844	),
    (angle:	-3.141592653589790	; val:	0.086266738334054	),
    (angle:	-2.617993877991490	; val:	0.145125825117650	),
    (angle:	-2.094395102393200	; val:	0.242610328729865	),
    (angle:	-1.570796326794900	; val:	0.398536815338387	),
    (angle:	-1.047197551196600	; val:	0.624887966296087	),
    (angle:	-0.523598775598299	; val:	0.877009640454779	),
    (angle:	0.000000000000000	; val:	1.000000000000000	),
    (angle:	0.523598775598298	; val:	0.877009640454780	),
    (angle:	1.047197551196600	; val:	0.624887966296087	),
    (angle:	1.570796326794900	; val:	0.398536815338387	),
    (angle:	2.094395102393190	; val:	0.242610328729865	),
    (angle:	2.617993877991490	; val:	0.145125825117650	),
    (angle:	3.141592653589790	; val:	0.086266738334054	),
    (angle:	3.665191429188090	; val:	0.051165011143844	),
    (angle:	4.188790204786390	; val:	0.030322266647950	),
    (angle:	4.712388980384690	; val:	0.017965132264752	),
    (angle:	5.235987755982990	; val:	0.010642829562681	),
    (angle:	5.759586531581290	; val:	0.006304766850925	));

procedure TTestHyperbolic.TestSech;
var
  i: integer;
begin
  for i := 0 to 23 do
    TestOk(Format('sech(%s)', [NForm(SechValues[i].angle)]), SechValues[i].val, eps11);
end;

procedure TTestInverseHyperbolic.TestASech;
var
  i: integer;
begin
  // defined 0 <= x <= 1 and a >= 0 only
  for i := 12 to 23 do
    TestOk(Format('asech(%s)', [NForm(SechValues[i].val)]), SechValues[i].angle, eps11);
end;

{ TTestVariableFunctions }

procedure TTestVariableFunctions.TestAvg;
begin
  TestOk('avg(3, 4, 5)', 4);
  TestOk('avg(3, 4)', 3.5);
  TestOk('avg(3)', 3);
end;

procedure TTestVariableFunctions.TestMax;
begin
  TestOk('max(6, 5, 4)', 6);
  TestOk('max(5, 4)', 5);
  TestOk('max(4)', 4);
  TestOk('max(4, -5)', 4);
  TestOk('max(-4, 5)', 5);
  TestOk('max(0, 0)', 0);
  TestOk('max(9, max(3, 4))', 9);
  TestOk('max(max(13, -9), max(3, 4))', 13);
end;

procedure TTestVariableFunctions.TestMin;
begin
  TestOk('min(2, 3, 4)', 2);
  TestOk('min(5, 4)', 4);
  TestOk('min(5)', 5);
  TestOk('min(4, 5)', 4);
  TestOk('min(4, -5)', -5);
  TestOk('min(-4, 5)', -4);
  TestOk('min(0, 0)', 0);
end;

procedure TTestVariableFunctions.TestPoly;
begin
  TestOk('poly(2, 1, 4, 3, -4)', 1 + 4*2 + 3*(2*2) - 4*(2*2*2));
  TestOk('poly(2, 0, 4, 0, -4)',  4*2 - 4*(2*2*2));
  TestOk('poly(2, 1, 4, 3)', 1 + 4*2 + 3*(2*2));
  TestOk('poly(2, 1, 4)', 1 + 4*2);
  TestOk('poly(2, 1)', 1 );
  TestBadParse('poly(2)', peTooFewParams);
end;

procedure TTestVariableFunctions.TestRecursion;
begin
  TestOk('avg(2, avg(1, 2, 3))', 2);
  TestOk('avg(0, avg(5, 5, 5, 5))', 2.5);
end;

procedure TTestVariableFunctions.TestSsq;
begin
  TestOk('ssq(3)', 9);
  TestOk('ssq(3, 4)', 25);
  TestOk('ssq(3, 4, 5)', 50);
end;

procedure TTestVariableFunctions.TestVar;
begin
  TestOk('var(3, 4, 5, 6)', 1 + 2/3, 1E-6);
  TestOk('var(3, 4, 5)', 1);
  TestOk('var(3, 4)', 0,5);
  TestBadParse('var(3)', peTooFewParams)
end;

procedure TTestVariableFunctions.TestVarp;
begin
  TestOk('varp(3, 4, 5, 6)', 1,25);
  TestOk('varp(3, 4, 5)', 2/3, 1E-6);
  TestOk('varp(3, 4)', 0,25);
  TestOk('varp(3)', 0);
end;

{ TestConstants }

procedure TestConstants.TestCombo;
begin
  TestOk('e*2', 2*exp(1));
  TestOk('ln(pi/e)', ln(pi/exp(1)));
end;

procedure TestConstants.Teste;
begin
  TestOk('e', exp(1));
end;

procedure TestConstants.TestPi;
begin
  TestOk('pi', pi);
end;

initialization
  // Recenser tous les cas de test avec le lanceur de test
  RegisterTest(TestConstants.Suite);
  RegisterTest(TestMathFunctions.Suite);
  RegisterTest(TestPowerFunctions.Suite);
  RegisterTest(TestParserAngle.Suite);
  RegisterTest(TestTrigonometricFunctions.Suite);
  RegisterTest(TestInverseTrigonometricFunctions.Suite);
  RegisterTest(TTestHyperbolic.Suite);
  RegisterTest(TTestInverseHyperbolic.Suite);
  RegisterTest(TTestVariableFunctions.Suite);
end.

